@@ -3,16 +3,16 @@ import os
 import sys
 
 from .arguments_parser import ArgumentsParser
-from . import arguments_handler
+from .arguments_handler import ArgumentsHandler
 
 
 def run():
     check_login_condition()
     parser = ArgumentsParser()
     args = parser.parse_arguments()
-    print('Namespace object before handling by this program: ' + str(args))
-    arguments_handler.check_and_reset_args(args)
-    arguments_handler.handle_arguments(args)
+    arguments_handler = ArgumentsHandler(args)
+    arguments_handler.check_and_preset_args()
+    arguments_handler.handle_arguments()
 
 
 def check_login_condition():
