@@ -3,7 +3,7 @@ from typing import List, Dict
 import gi
 import psutil
 
-from utils import snapd_workaround
+from ..utils import snapd_workaround
 
 gi.require_version('Wnck', '3.0')
 from gi.repository import Wnck
@@ -43,6 +43,6 @@ def test_get_desktop_file():
 
 
 def test_launch_app():
-    launched = snapd_workaround.Snapd().launch(['authy'], False, False)
+    launched = snapd_workaround.Snapd().launch_app(['authy'], False, False)
     assert launched is True
 
